/*
 * To do:
 *	find a purpose for this...
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "amd64.h"

/*
 * Address Space Map.
 * Low duty cycle.
 */
typedef struct Asm Asm;
typedef struct Asm {
	uintmem	addr;
	uintmem	size;
	int	type;
	int	location;
	Asm*	next;
} Asm;

enum {
	AsmNone		= 0,
	AsmMemory,
	AsmReserved,
	AsmAcpireclaim,
	AsmAcpinvs,
	AsmUnusable,
	AsmDisable,

	AsmPci,					/* invented */
	AsmLast			= AsmPci,
};

static Lock asmlock;
static Asm asmarray[100] = {
	{ 0, ~0, AsmNone, nil, },
};
static int asmindex = 1;
static Asm* asmlist = &asmarray[0];
static Asm* asmfreelist;
static char *tname[AsmLast+1] = {
	"none",
	"mem",
	"res",
	"reclaim",
	"nvs",
	"unusable",
	"disable",

	"pcimap",
};

static long
asmread(Chan*, void *v, long n, vlong off)
{
	char *s, *p, *e;
	Asm *a;

	s = smalloc(16*1024);
	e = s + 16*1024;
	p = s;
	for(a = asmlist; a != nil; a = a->next)
		p = seprint(p, e, "%.8s	%#.16p	%#.16p\n",
			tname[a->type], a->addr, a->addr+a->size);
	n = readstr(off, v, n, s);
	free(s);
	return n;
}

void
asmdump(void)
{
	Asm *a;

	for(a = asmlist; a != nil; a = a->next)
		print("%.8s	%#.16p	%#.16p\n",
			tname[a->type], a->addr, a->addr+a->size);
}

static Asm*
asmnew(uintmem addr, uintmem size, int type)
{
	Asm * asm;

	if(asmfreelist != nil){
		asm = asmfreelist;
		asmfreelist = asm->next;
		asm->next = nil;
	}
	else{
		if(asmindex >= nelem(asmarray))
			return nil;
		asm = &asmarray[asmindex++];
	}
	asm->addr = addr;
	asm->size = size;
	asm->type = type;

	return asm;
}

int
asmfree(uintmem addr, uintmem size, int type)
{
	Asm *np, *pp, **ppp;

	DBG("asmfree: %#P@%#P, type %d\n", size, addr, type);
	if(size == 0)
		return 0;
	if(type > AsmLast){
		print("asmfree: bad type from %#p\n", getcallerpc(&addr));
		return -1;
	}

	lock(&asmlock);

	/*
	 * Find either a map entry with an address greater
	 * than that being returned, or the end of the map.
	 */
	pp = nil;
	ppp = &asmlist;
	for(np = *ppp; np != nil && np->addr <= addr; np = np->next){
		pp = np;
		ppp = &np->next;
	}

	if((pp != nil && pp->addr+pp->size > addr)
	|| (np != nil && addr+size > np->addr)){
		unlock(&asmlock);
		DBG("asmfree: overlap %#Px@%#P, type %d\n", size, addr, type);
		return -1;
	}

	if(pp != nil && pp->type == type && pp->addr+pp->size == addr){
		pp->size += size;
		if(np != nil && np->type == type && addr+size == np->addr){
			pp->size += np->size;
			pp->next = np->next;

			np->next = asmfreelist;
			asmfreelist = np;
		}

		unlock(&asmlock);
		return 0;
	}

	if(np != nil && np->type == type && addr+size == np->addr){
		np->addr -= size;
		np->size += size;

		unlock(&asmlock);
		return 0;
	}

	if((pp = asmnew(addr, size, type)) == nil){
		unlock(&asmlock);
		DBG("asmfree: losing %#P@%#P, type %d\n", size, addr, type);
		return -1;
	}
	*ppp = pp;
	pp->next = np;

	unlock(&asmlock);

	return 0;
}

uintmem
asmalloc(uintmem addr, uintmem size, int type, int align)
{
	uintmem a, o;
	Asm *asm, *pp;

	DBG("asmalloc: %#.16llux@%#.16llux, %s\n", size, addr, tname[type]);
//	print("asmalloc: %#.16llux@%#.16llux, %s\n", size, addr, tname[type]);
	lock(&asmlock);
	for(pp = nil, asm = asmlist; asm != nil; pp = asm, asm = asm->next){
		if(asm->type != type)
			continue;
		a = asm->addr;

		if(addr != 0){
			/*
			 * A specific address range has been given:
			 *   if the current map entry is greater then
			 *   the address is not in the map;
			 *   if the current map entry does not overlap
			 *   the beginning of the requested range then
			 *   continue on to the next map entry;
			 *   if the current map entry does not entirely
			 *   contain the requested range then the range
			 *   is not in the map.
			 * The comparisons are strange to prevent
			 * overflow.
			 */
			if(a > addr)
				break;
			if(asm->size < addr - a)
				continue;
			if(addr - a > asm->size - size)
				break;
			a = addr;
		}

		if(align > 0)
			a = ((a+align-1)/align)*align;
		if(asm->addr+asm->size-a < size)
			continue;
if(type == AsmMemory)print("	asm %#.16llux@%#.16llux\n", size, addr);

		o = asm->addr;
		asm->addr = a+size;
		asm->size -= a-o+size;
		if(asm->size == 0){
			if(pp != nil)
				pp->next = asm->next;
			asm->next = asmfreelist;
			asmfreelist = asm;
		}

		unlock(&asmlock);
if(type == AsmMemory)print("	asm %#.16llux@%#.16llux\n", a-o, o);
		if(o != a)
			asmfree(o, a-o, type);
		return a;
	}
	unlock(&asmlock);

print("asmalloc: fail addr %p size %p type %s align %d\n", addr, size, tname[type], align);
asmdump();

	return 0;
}

static void
asminsert(uintmem addr, uintmem size, int type)
{
	if(type == AsmNone || asmalloc(addr, size, AsmNone, 0) == 0)
		return;
	if(asmfree(addr, size, type) == 0)
		return;
	asmfree(addr, size, 0);
}

void
asminit(void)
{
	sys->pmstart = ROUNDUP(PADDR(end), PGSZ);
	sys->pmend = sys->pmstart;
	asmalloc(0, sys->pmstart, AsmNone, 0);
	addarchfile("asm", 0444, asmread, nil);
}

/*
 * Notes:
 * asmmapinit called from options.c:/^e820
 * subject to change; the numerology here is probably suspect.
 */
void
asmmapinit(uintmem addr, uintmem size, int type)
{
	switch(type){
	default:
		asminsert(addr, size, type);
		break;
	case AsmMemory:
		/*
		 * Adjust things for the peculiarities of this
		 * architecture.
		 * Sys->pmend is the largest physical memory address found,
		 * there may be gaps between it and sys->pmstart, the range
		 * and how much of it is occupied, might need to be known
		 * for setting up allocators later.
		 */
		if(addr < 1*MiB || addr+size < sys->pmstart)
			break;
		if(addr < sys->pmstart){
			size -= sys->pmstart - addr;
			addr = sys->pmstart;
		}
		asminsert(addr, size, type);
		if(addr+size > sys->pmend)
			sys->pmend = addr+size;
		break;
	}
}

void*
asmbootalloc(usize size)
{
	uintptr va;

	assert(sys->vmunused+size <= sys->vmunmapped);
	va = sys->vmunused;
	sys->vmunused += size;
	memset(UINT2PTR(va), 0, size);
	return UINT2PTR(va);
}

static PTE
asmwalkalloc(usize size)
{
	uintmem pa;

	assert(size == PTSZ && sys->vmunused+size <= sys->vmunmapped);

	if(!ALIGNED(sys->vmunused, PTSZ)){
		DBG("asmwalkalloc: %ulld wasted\n",
			ROUNDUP(sys->vmunused, PTSZ) - sys->vmunused);
		sys->vmunused = ROUNDUP(sys->vmunused, PTSZ);
	}
	if((pa = mmuphysaddr(sys->vmunused)) != ~0)
		sys->vmunused += size;

	return pa;
}

static void
confcrap(void)
{
	int n;
	Asm* asm;
	uintmem lo, hi;

	n = 0;
	for(asm = asmlist; asm != nil; asm = asm->next){
		/*
		 * this is only used for port/devproc.c: Qmem cases
		 * convert that to use physmem and we can drop conf.mem.
		 */
		if(asm->type != AsmMemory || n >= nelem(conf.mem))
			continue;
		lo = ROUNDUP(asm->addr, PGSZ);
		hi = ROUNDDN(asm->addr+asm->size, PGSZ);
		conf.mem[n].base = lo;
		conf.mem[n].npage = (hi - lo)/PGSZ;
		sys->npages += conf.mem[n].npage;
		sys->upages += conf.mem[n].npage;
		print("->cm %d: addr %#llux npage %lud\n",
			n, conf.mem[n].base, conf.mem[n].npage);
		n++;
	}
	print("npage %llud upage %llud kpage %llud\n",
		sys->npages, sys->upages, sys->kpages);
}

void
asmmeminit(void)
{
	int i, l, npg[NPGSZ];
	Asm* asm;
	PTE *pte, *pml4;
	uintptr va;
	uintmem hi, lo, sum, mem, nextmem, pa;

	assert(!((sys->vmunmapped|sys->vmend) & m->pgszmask[1]));

	/* make sure that memory from vmend to TMFM is not in the map */
	if((pa = mmuphysaddr(sys->vmunused)) == ~0)
		panic("asmmeminit 1");
	pa += sys->vmunmapped - sys->vmunused;
	mem = asmalloc(pa, sys->vmend - sys->vmunmapped, AsmMemory, 0);
	if(mem != pa)
		panic("asmmeminit 2");
	DBG("pa %#llux mem %#llux\n", pa, mem);

	/* assume already 2MiB aligned*/
	assert(ALIGNED(sys->vmunmapped, 2*MiB));
	pml4 = UINT2PTR(m->pml4->va);
	while(sys->vmunmapped < sys->vmend){
		l = mmuwalk(pml4, sys->vmunmapped, 1, &pte, asmwalkalloc);
		DBG("%#p l %d\n", sys->vmunmapped, l);
		*pte = pa|PtePS|PteRW|PteP|PteG;
		sys->vmunmapped += 2*MiB;
		pa += 2*MiB;
		sys->kpages += 2*MiB/PGSZ;
		sys->npages += 2*MiB/PGSZ;
	}

	memset(npg, 0, sizeof npg);
	sum = 0;
	for(asm = asmlist; asm != nil; asm = asm->next){
		if(asm->type != AsmMemory)
			continue;
		va = KSEG2+asm->addr;
		print("asm: mem %#P %#P size %P\n",
			asm->addr, asm->addr+asm->size, asm->size);

		lo = asm->addr;
		hi = asm->addr+asm->size;
		sum += hi - lo;
		/* Convert a range into pages */
		for(mem = lo; mem < hi; mem = nextmem){
			nextmem = (mem + PGLSZ(0)) & ~m->pgszmask[0];

			/* Try large pages first */
			for(i = m->npgsz - 1; i >= 0; i--){
				if((mem & m->pgszmask[i]) != 0)
					continue;
				if(mem + PGLSZ(i) > hi)
					continue;
				/* This page fits entirely within the range. */
				/* Mark it a usable */
				if((l = mmuwalk(pml4, va, i, &pte, asmwalkalloc)) < 0)
					panic("asmmeminit 3");

				*pte = mem|PteRW|PteP|PteG;
				if(l > 0)
					*pte |= PtePS;

				nextmem = mem + PGLSZ(i);
				va += PGLSZ(i);
				npg[i]++;

				break;
			}
		}
	}
	print("%d %d %d\n", npg[0], npg[1], npg[2]);
	confcrap();
	sys->ialloc = (sum/4)*PGSZ;		/* should be in physmem */
}

void
asmumeminit(void)
{
	Asm *asm;
	extern void physallocdump(void);

	for(asm = asmlist; asm != nil; asm = asm->next){
		if(asm->type != AsmMemory)
			continue;
		physinit(asm->addr, asm->size);
	}
	physallocdump();
}
